# SPDX-FileCopyrightText: 2025-present Minami Ueda
#
# SPDX-License-Identifier: MIT

# DO NOT MODIFY THIS FILE DIRECTLY.
# To make a new release, use the GitHub Actions workflow (release.yml), instead of modifying the version values below directly.
# Detailed documentation are available in DEVELOP.md.

__version__ = "0.3.0"
